﻿/* Copyright � 2019 Roesner P/L

GENERAL INFO
Plain c executable for particle motion processing. For rotating disc granular spreader applications.
This code reads an input file, containing all user parameters. Results are written to an ouput file.

Processing Part1 handles particle motion along the spinner vane. Acceleration is calculated repeatedly
at user specified intervals. This breaks the vane travel into segments. Exit angle (about the disc axis)
and velocity are calculated.

Processing Part2 handles the airborne trajectory to determine landing coordinates.

This is built for speed. Inline calcs are used in preference of library calls where possible. File
read/write operations are batched at the start and end of process only.

All coordinates and angles are referenced to the center of the disc.

Vane iterations, 100 000 particles at fortran resolution = 5.5sec

*/

#define _WIN32_WINNT 0x0500
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <windows.h>

void readfile(void);
void readcsv(void);
void prepValues(void);
void calcVanePath(void);
void calcFlightPath(void);
void resetVariables(void);
void writecsv(void);


//input variables
double discRpm = 0;
double exitRadius = 0;
double exitHeight = 0;
double vaneIncline = 0;
double vaneRake = 0;
double particleDia = 0;
double particleMass = 0;
double xEntryPoint = 0;
double yEntryPoint = 0;
double vaneFriction = 0;
double flightDrag = 0;
double vaneInterval = 0;
double flightInterval = 0;
int viewResults = 0;

//variables used in calcs, ram is cheap and fast, use it
double halfPi = 1.5707963267948966;
double pi = 3.141592653589793238;
double oneAndHalfPi = 4.7123889803846897;
double twoPi = 6.2831853071795862;
double gravity = 9.80665;
double entryRadius = 0;
double entryAngle = 0;
double locationRadius = 0;
double discRadPerSec = 0;
double vaneInclineRad = 0;
double vaneRakeRad = 0;
double vaneAccel = 0;
double contactDistance = 0;
double contactTime = 0;
double exitVelocity = 0;
double contactRotation = 0;
double exitAngle = 0;
double particleRadVelocity = 0;     //radial velocity outwards along the vane
double tanVelocity = 0;             //tangential velocity at point of exit
double xRange = 0;                  //horizontal flight distance x
double yRange = 0;                  //horizontal flight distance y
double xExitPoint = 0;              //leaving the vane
double yExitPoint = 0;              //leaving the vane
double xLandPoint = 0;              //touchdown
double yLandPoint = 0;              //touchdown
double xExitVelocity = 0;				//leaving the vane
double yExitVelocity = 0;				//leaving the vane
double zExitVelocity = 0;				//leaving the vane


double var1 = 0;
double var2 = 0;
double var3 = 0;


// Stuff for looping multiple particles

// Helpers
int numParticles;
#define MAX_NUM_PARTICLES 5000	// This is arbitrary, you can increase this to increase max number of particles allowable

// Inputs
double xEntryPoints[MAX_NUM_PARTICLES];
double yEntryPoints[MAX_NUM_PARTICLES];
double particleMasses[MAX_NUM_PARTICLES];
double particleDiameters[MAX_NUM_PARTICLES];

// Outputs
double xLandPoints[MAX_NUM_PARTICLES];
double yLandPoints[MAX_NUM_PARTICLES];
double exitAngles[MAX_NUM_PARTICLES];
double xExitVelocities[MAX_NUM_PARTICLES];
double yExitVelocities[MAX_NUM_PARTICLES];
double zExitVelocities[MAX_NUM_PARTICLES];


void main() 
{
	//set the console window position
	HWND hWnd = GetConsoleWindow();
	MoveWindow(hWnd, 20, 20, 800, 900, TRUE);

	readfile();
	readcsv();

	printf("\n\nCALCULATING...");
	for(int i = 0; i < numParticles; i++) 
	{		
		xEntryPoint = xEntryPoints[i];
		yEntryPoint = yEntryPoints[i];
		particleMass = particleMasses[i];
		particleDia = particleDiameters[i];

		prepValues();
		calcVanePath();
		calcFlightPath();

		exitAngles[i] = exitAngle;
		xLandPoints[i] = xLandPoint;
		yLandPoints[i] = yLandPoint;
		xExitVelocities[i] = xExitVelocity;
		yExitVelocities[i] = yExitVelocity;
		zExitVelocities[i] = zExitVelocity;

		resetVariables();
	}
	
	writecsv();

	printf("\nFINISHED\n\nHit ENTER");
	getchar();
}

void resetVariables(void) {
	particleRadVelocity = 0;
}

void prepValues(void) {
	//calc distance to center of disc, hypotenuse of xy dropPoint, if x or y is negative make them positive
	var1 = xEntryPoint;
	if (var1 < 0) { var1 = -var1; }

	var2 = yEntryPoint;
	if (var2 < 0) { var2 = -var2; }

	var3 = (var1 * var1) + (var2 * var2);
	entryRadius = sqrt(var3);
	//printf("\n\nparticle entry point radius: %1.10f m\n", entryRadius);
	locationRadius = entryRadius;

	//calc angle of entry point, see math atan2 for details
	//if y is negative, it returns negative value, so we flip it to produce positive radians
	//3oclock = 0, radians increase all the way round in counterclockwise rotation
	entryAngle = atan2(yEntryPoint, xEntryPoint);
	if (entryAngle < 0) {
		entryAngle = twoPi + entryAngle;
	}
	//printf("particle entry angle: %1.10f rad\n", entryAngle);

	//angular velocity in radian/sec
	discRadPerSec = (twoPi * discRpm) / 60;

	//vane angles to radians
	vaneInclineRad = (vaneIncline * pi) / 180;
	vaneRakeRad = (vaneRake * pi) / 180;
}

void calcVanePath(void) {

	double segments = 0;
	double distance1 = 0;
	double distance2 = 0;
	double segmentDistance = 0;
	double particleVelIncrease = 0;

	while (locationRadius < exitRadius) {

		segments++;
		//printf("\nsegment number: %1.0f\n", segments);
		//printf("start radius: %1.10f\n", locationRadius);

		//particle radial acceleration on a basic vane (0 deg vane incline, 0 deg vane rake)
		//acc = radius * angular velocity squared
		vaneAccel = locationRadius * discRadPerSec * discRadPerSec;
		//printf("acceleration for this segment: %1.10f\n", vaneAccel);

		//apply vane incline and rake angles
		vaneAccel = vaneAccel * cos(vaneRakeRad) * cos(vaneInclineRad);
		//printf("vane acceleration incl angles: %1.10f\n", vaneAccel);

		//apply gravity contribution
		vaneAccel = vaneAccel - (cos(vaneRakeRad) * sin(vaneInclineRad) * gravity);
		//printf("vane acceleration incl gravity: %1.10f\n", vaneAccel);

		//apply vane incline friction, due to gravity and centrifugal force
		vaneAccel = vaneAccel - vaneFriction * ((cos(vaneInclineRad) * gravity) + (sin(vaneInclineRad) * vaneAccel));

		//apply vane rake friction, due to centrifugal force and gravity
		vaneAccel = vaneAccel - vaneFriction * ((sin(vaneRakeRad) * vaneAccel) - (sin(vaneRakeRad) * sin(vaneInclineRad) * gravity));
		//printf("vane acceleration incl friction: %1.10f\n", vaneAccel);

		//distance it will travel due to it's initial velocity during this segment
		distance1 = particleRadVelocity * vaneInterval;
		//distance it will travel due to it's acceleration during this segment
		//distance travelled = 1/2 acceleration * time squared
		distance2 = vaneInterval * vaneInterval * (vaneAccel / 2);
		segmentDistance = distance1 + distance2;
		//printf("distance travelled for this segment: %1.10f\n", segmentDistance);
		locationRadius += segmentDistance;

		//accumulate the particle velocity increase
		particleVelIncrease = vaneInterval * vaneAccel;
		particleRadVelocity += particleVelIncrease;
		//printf("velocity increase: %1.10f\n", particleVelIncrease);
		//printf("velocity: %1.10f\n", particleRadVelocity);

	}

	//final disc values
	contactTime = segments * vaneInterval;
	contactDistance = exitRadius - entryRadius;
	contactRotation = discRadPerSec * contactTime;
	exitAngle = entryAngle + contactRotation;
	//if exit angle has rolled around past 3oclock, subtract 2pi radians
	if (exitAngle > twoPi) {
		exitAngle = exitAngle - twoPi;
	}
	tanVelocity = discRadPerSec * exitRadius;

	exitVelocity = sqrt((tanVelocity * tanVelocity) + (particleRadVelocity * particleRadVelocity));

	/*
	printf("\n\nVANE CONTACT RESULTS\n");
	printf("steps on the vane: %1.0f\n", segments);
	printf("distance travelled on the vane: %1.10f m\n", contactDistance);
	printf("time travelled on the vane: %1.10f s\n", contactTime);
	printf("disc rotation during vane contact: %1.10f rad\n", contactRotation);
	printf("disc exit angle: %1.10f rad\n", exitAngle);
	printf("particle tangential velocity: %1.10f m/s\n", tanVelocity);
	printf("particle radial velocity: %1.10f m/s\n", particleRadVelocity);
	printf("particle exit velocity: %1.10f m/s\n\n", exitVelocity);
	*/
}

void calcFlightPath(void) {

	//printf("\nTRAJECTORY RESULTS\n");

	//split into vertical and horizontal velocity
	double verticalVel = particleRadVelocity * sin(vaneInclineRad);
	double horizontalRadVel = particleRadVelocity * cos(vaneInclineRad);

	double horizontalVel = sqrt((tanVelocity * tanVelocity) + (horizontalRadVel * horizontalRadVel));
	//printf("vertical velocity: %1.10f m/s\n", verticalVel);
	//printf("horizontal velocity: %1.10f m/s\n", horizontalVel);

	//flight path viewed from above, comes from radial and tangential velocity
	//add this to the angle of the vane at exit time
	double flightAngle = atan2(horizontalRadVel, tanVelocity);
	flightAngle = flightAngle + exitAngle;
	//if it has rolled around past 3oclock, subtract 2pi radians
	if (flightAngle > twoPi) {
		flightAngle = flightAngle - twoPi;
	}
	//printf("xy plane flight angle: %1.10f rad\n", flightAngle);

	//convert to cartesian coordinates
	xExitVelocity = horizontalVel * cos(flightAngle);
	yExitVelocity = horizontalVel * sin(flightAngle);
	zExitVelocity = verticalVel;

	//calc flight time
	double upFlightTime = verticalVel / gravity;
	//printf("upwards flight time: %1.10f s\n", upFlightTime);
	double maxHeight = (verticalVel * verticalVel) / (2 * gravity);
	maxHeight = maxHeight + exitHeight;
	//printf("max height: %1.10f m\n", maxHeight);
	double downFlightTime = sqrt(2 * maxHeight / gravity);
	//printf("down flight time: %1.10f s\n", downFlightTime);
	double totalFlightTime = upFlightTime + downFlightTime;
	//printf("total flight time: %1.10f s\n", totalFlightTime);

	double flightTimer = 0;
	double segments = 0;
	double range = 0;
	double distance1 = 0;
	double distance2 = 0;
	double segmentDistance = 0;
	double velocityDecrease = 0;
	double acceleration = 0;

	double airDensity = 1.225;
	double particleRadius = particleDia / 2;
	double particleArea = pi * particleRadius * particleRadius;


	//calc the horizontal distance for each segment in flight
	//distance travelled = distance due to initial velocity plus distance due to acceleration
	//in this case the acceleration is negative (deceleration)
	while (flightTimer < totalFlightTime) {

		segments++;
		flightTimer += flightInterval;

		//horizontal force on the particle in flight, due to air drag
		//force = 0.5 * air density * velocity^2 * frontal area * drag coefficient
		//now find acceleration
		//force = mass * acceleration
		//acceleration = (air density * velocity^2 * frontal area * drag coefficent) / (2 * mass)

		acceleration = (airDensity * horizontalVel * horizontalVel * particleArea * flightDrag) / (2 * particleMass);
		//printf("acceleration: %1.10f\n", acceleration);

		//distance it will travel due to it's initial velocity during this segment
		distance1 = horizontalVel * flightInterval;
		//distance it will travel due to it's acceleration during this segment
		//distance travelled = 1/2 acceleration * time squared
		distance2 = flightInterval * flightInterval * (acceleration / 2);
		segmentDistance = distance1 - distance2;


		range += segmentDistance;

		//calc new velocity due to deceleration
		velocityDecrease = flightInterval * acceleration;
		horizontalVel -= velocityDecrease;
		//printf("flight velocity decrease: %1.10f\n", velocityDecrease);
		//printf("flight velocity: %1.10f\n", horizontalVel);


	}

	//printf("steps in flight: %1.10f\n", segments);
	//printf("flight range: %1.10f m\n", range);



	//we want the xy landing point from the center of the spinner disc
	//so we need to add the xy flight distances to the xy exit point
	//first get the xy flight distances based on the the flight angle
	//cos and sin of flight angle automatically return correct positive or negative coordinate
	xRange = range * cos(flightAngle);
	yRange = range * sin(flightAngle);

	//now get the xy points where the particle exit occurred
	xExitPoint = exitRadius * cos(exitAngle);
	yExitPoint = exitRadius * sin(exitAngle);
	//printf("xExitPoint: %1.10f m\n", xExitPoint);
	//printf("yExitPoint: %1.10f m\n", yExitPoint);

	//sum these to get the landing point from the center of the disc
	xLandPoint = xExitPoint + xRange;
	yLandPoint = yExitPoint + yRange;
	//printf("xLandPoint: %1.10f m\n", xLandPoint);
	//printf("yLandPoint: %1.10f m\n", yLandPoint);
}

void readfile(void) {
	//these are pointers to strings
	char *inputKey;
	char *inputVal;
	char *delimiter = "=";

	//read the input file, absolute or realtive path
	//FILE *file = fopen("D:/inputFile.txt", "r");
	FILE *file = fopen("./inputFile.txt", "r");

	if (file != NULL) {
		printf("READ INPUT FILE\n");
		//cycle through each line, read characters into an array, check for key-value pairs
		char line[128];
		while (fgets(line, sizeof line, file) != NULL) {
			//grab the string before the delimiter, use stringToken
			inputKey = strtok(line, delimiter);
			//if string compare is zero, the key matches
			//get the string after the delimiter, load it to the variable
			if (strcmp(inputKey, "discRpm") == 0) {
				inputVal = strtok(NULL, delimiter);
				discRpm = strtod(inputVal, NULL);
				printf("loaded discRpm: %1.10f\n", discRpm);
			}
			else if (strcmp(inputKey, "exitRadius") == 0) {
				inputVal = strtok(NULL, delimiter);
				exitRadius = strtod(inputVal, NULL);
				printf("loaded exitRadius: %1.10f\n", exitRadius);
			}
			else if (strcmp(inputKey, "exitHeight") == 0) {
				inputVal = strtok(NULL, delimiter);
				exitHeight = strtod(inputVal, NULL);
				printf("loaded exitHeight: %1.10f\n", exitHeight);
			}
			else if (strcmp(inputKey, "vaneIncline") == 0) {
				inputVal = strtok(NULL, delimiter);
				vaneIncline = strtod(inputVal, NULL);
				printf("loaded vaneIncline: %1.10f\n", vaneIncline);
			}
			else if (strcmp(inputKey, "vaneRake") == 0) {
				inputVal = strtok(NULL, delimiter);
				vaneRake = strtod(inputVal, NULL);
				printf("loaded vaneRake: %1.10f\n", vaneRake);
			}
			/*
				else if (strcmp(inputKey, "particleDia") == 0) {
				inputVal = strtok(NULL, delimiter);
				particleDia = strtod(inputVal, NULL);
				printf("loaded particleDia: %1.10f\n", particleDia);
			}
			else if (strcmp(inputKey, "particleMass") == 0) {
				inputVal = strtok(NULL, delimiter);
				particleMass = strtod(inputVal, NULL);
				printf("loaded particleMass: %1.10f\n", particleMass);
			}
			else if (strcmp(inputKey, "xEntryPoint") == 0) {
				inputVal = strtok(NULL, delimiter);
				xEntryPoint = strtod(inputVal, NULL);
				printf("loaded xEntryPoint: %1.10f\n", xEntryPoint);
			}
			else if (strcmp(inputKey, "yEntryPoint") == 0) {
				inputVal = strtok(NULL, delimiter);
				yEntryPoint = strtod(inputVal, NULL);
				printf("loaded yEntryPoint: %1.10f\n", yEntryPoint);
			}
			*/
			else if (strcmp(inputKey, "vaneFriction") == 0) {
				inputVal = strtok(NULL, delimiter);
				vaneFriction = strtod(inputVal, NULL);
				printf("loaded vaneFriction: %1.10f\n", vaneFriction);
			}
			else if (strcmp(inputKey, "flightDrag") == 0) {
				inputVal = strtok(NULL, delimiter);
				flightDrag = strtod(inputVal, NULL);
				printf("loaded flightDrag: %1.10f\n", flightDrag);
			}
			else if (strcmp(inputKey, "vaneInterval") == 0) {
				inputVal = strtok(NULL, delimiter);
				vaneInterval = strtod(inputVal, NULL);
				printf("loaded vaneInterval: %1.10f\n", vaneInterval);
			}
			else if (strcmp(inputKey, "flightInterval") == 0) {
				inputVal = strtok(NULL, delimiter);
				flightInterval = strtod(inputVal, NULL);
				printf("loaded flightInterval: %1.10f\n", flightInterval);
			}
			else if (strcmp(inputKey, "viewResults") == 0) {
				inputVal = strtok(NULL, delimiter);
				if (strcmp(inputVal, "on") == 0) {
					viewResults = 1;
					printf("view results on\n");
				}
				else {
					viewResults = 0;
					printf("view results off\n");
				}
			}
		}
	}

	//close file
	fclose(file);
}

void readcsv(void) {
	//read the input file, absolute or realtive path
	//FILE *file = fopen("D:/input.csv", "r");
	FILE *file = fopen("./input.csv", "r");

	char buf[1024];

	if (!file) {
		printf("\n\nCANNOT OPEN CSV\n");
		return;
	}

	printf("\n\nREAD INPUT CSV\n");

	int row_count = 0;
	int field_count = 0;
	int pointer = 0;

	while (fgets(buf, 1024, file)) {
		field_count = 0;
		row_count++;

		// Skip column headers of csv
		if (row_count == 1) {
			continue;
		}

		char *field = strtok(buf, ",");
		while (field) {
			if (field_count == 0) {
				xEntryPoints[pointer] = strtod(field, NULL);
			}
			else if (field_count == 1) {
				yEntryPoints[pointer] = strtod(field, NULL);
			}
			else if (field_count == 2) {
				particleMasses[pointer] = strtod(field, NULL);
			}
			else if (field_count == 3) {
				particleDiameters[pointer] = strtod(field, NULL);
			}

			field = strtok(NULL, ",");

			field_count++;
		}

		/*
		printf("loaded xEntryPoint: %1.10f\n", xEntryPoints[pointer]);
		printf("loaded yEntryPoint: %1.10f\n", yEntryPoints[pointer]);
		printf("loaded mass: %1.10f\n", particleMasses[pointer]);
		printf("loaded diameter: %1.10f\n", particleDiameters[pointer]);
		printf("\n");
		*/

		pointer++;

		if (pointer == MAX_NUM_PARTICLES) {
			break;
		}
	}

	numParticles = pointer;
	printf("Number of particles loaded: %d\n", numParticles);

	//close file
	fclose(file);
}

void writecsv(void) {
	//write the output file, absolute or relative path
	//FILE *file = fopen("D:/output.csv", "w");
	FILE *file = fopen("./output.csv", "w");

	// Write column headers
	fputs("ID,Mass (kg),Diameter (m),x Land Point (m),y Land Point (m),Exit Angle (rad),x Exit Velocity (m/s),y Exit Velocity (m/s),z Exit Velocity (m/s)\n", file);

	// Write data
	for(int i = 0; i < numParticles; i++) {
		fprintf(file, "%04d", i);
		fputs(",", file);
		fprintf(file, "%1.10f", particleMasses[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", particleDiameters[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", xLandPoints[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", yLandPoints[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", exitAngles[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", xExitVelocities[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", yExitVelocities[i]);
		fputs(",", file);
		fprintf(file, "%1.10f", zExitVelocities[i]);
		fputs("\n", file);
	}

	//close file
	fclose(file);
}